/*
 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <curses.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <modules.h>
#include <util.h>
#include <dstring.h>
#include <interp.h>
#include <limits.h>
#include "version.h"
//// \author[Artem V. Andreev]
//// \copyright[2001, 2002 Artem V. Andreev. See \ref[sl(1)] for details]
/// \section[NAME]
// \em[ncurses] -- a module for Tensile (see \ref[sl(1)])

/// \section[OVERVIEW]
//// This module provides interface to *nix \em[(n)curses] library.
//// It does not do direct binding, but rather uses Tensile streams interface

typedef struct int_table
{
	const char *name;
	int val;
} int_table;

static module *this_mod;
extern procstack *get_pstack(void);

void ncurses_LTX_init(module *this)
{
	this_mod = this;
}

//// \section[FUNCTIONS]
//// sysvar
//// \em[ncurses] module adds the following system variables (read-only):
//// \item \em[term:columns] -- the width of the screen
//// \item \em[term:rows] -- the height of the screen
//// \endlist
static void bi_sysvar(chunk *current, interpreter *i, value *args)
{
	instruction_t *saved = get_ma_data(this_mod, i);
	if(!ds_comparestr(args->val, "term:columns", NULL, NULL))
	{
		if(next_value(args))
			recover_error("'term:columns' is read-only");
		ds_append(get_pstack()->state->buffer, ds_fromint(COLS, 10, 0));
	}
	else if(!ds_comparestr(args->val, "term:rows", NULL, NULL))
	{
		if(next_value(args))
			recover_error("'term:rows' is read-only");
		ds_append(get_pstack()->state->buffer, ds_fromint(LINES, 10, 0));
	}
	else (saved[1])(current, i, args);
}

static int name_to_key(const char *key);
//// feature
//// \em[ncurses] adds the following features:
//// \item \em[::term:colors] true if the terminal supports coloring
//// \item \em[::term:mouse] true if the terminal has mouse and the underlying library
//// supports it
//// \item \em[::term:key:]keyname true if the terminal has the named \em[keyname]
//// \endlist 
static void bi_feature(chunk *current, interpreter *i, value *args)
{
	instruction_t *saved = get_ma_data(this_mod, i);
	char *name = DS_BODY(args->val);

	if(!strcmp(name, "::term:colors"))
	{
		if(has_colors())
			ds_appendstr(get_pstack()->state->buffer, "true");
	}
#ifdef NCURSES_MOUSE_VERSION
	else if(!strcmp(name, "::term:mouse"))
	{
		mmask_t saved;
		if(mousemask(ALL_MOUSE_EVENTS, &saved))
		{
			mousemask(saved, NULL);
			ds_appendstr(get_pstack()->state->buffer, "true");
		}
	}
#endif
#ifdef NCURSES_VERSION
	else if(!strncmp(name, "::term:key:", 11))
	{
		int key = name_to_key(name + 11);
		if(key > 0 && has_key(key))
			ds_appendstr(get_pstack()->state->buffer, "true");
	}
#endif
	else (**saved)(current, i, args);
}

#ifdef NCURSES_MOUSE_VERSION

#define BUTTON_PATTERN(event) switch(*++mstr) \
{ \
case '1': \
mask |= BUTTON1_##event; \
break; \
case '2': \
mask |= BUTTON2_##event; \
break; \
case '3': \
mask |= BUTTON3_##event; \
break; \
case '4': \
mask |= BUTTON4_##event; \
break; \
default: \
	recover_error("ncurses: unknown mouse button %c", mstr[1]); \
}


static mmask_t convert_mouse_mask(char *mstr)
{
	mmask_t mask = 0;
	if(!*mstr)
		return 0;
	if(!strcmp(mstr, "all") || (*mstr == '*' && !mstr[1]))
		return ALL_MOUSE_EVENTS;
	for(; *mstr; mstr++)
	{
		switch(*mstr)
		{
		case 'p':
			BUTTON_PATTERN(PRESSED);
			break;
		case 'r':
			BUTTON_PATTERN(RELEASED);
			break;
		case 'c':
			BUTTON_PATTERN(CLICKED);
			break;
		case 'd':
			BUTTON_PATTERN(DOUBLE_CLICKED);
			break;
		case 't':
			BUTTON_PATTERN(TRIPLE_CLICKED);
			break;
		case 'S':
			mask |= BUTTON_SHIFT;
			break;
		case 'A':
			mask |= BUTTON_ALT;
			break;
		case 'C':
			mask |= BUTTON_CTRL;
			break;
		case 'm':
			mask |= REPORT_MOUSE_POSITION;
			break;
		default:
			recover_error("ncurses: unrecognized mouse mask '%s'", mstr);
		}
	}
	return mask;
}

static dstring *stringize_mouse_mask(mmask_t mask)
{
	static int_table events[] = {{"p1", BUTTON1_PRESSED},
								 {"r1", BUTTON1_RELEASED},
								 {"c1", BUTTON1_CLICKED},
								 {"d1", BUTTON1_DOUBLE_CLICKED},
								 {"t1", BUTTON1_TRIPLE_CLICKED},
								 {"p2", BUTTON2_PRESSED},
								 {"r2", BUTTON2_RELEASED},
								 {"c2", BUTTON2_CLICKED},
								 {"d2", BUTTON2_DOUBLE_CLICKED},
								 {"t2", BUTTON2_TRIPLE_CLICKED},
								 {"p3", BUTTON3_PRESSED},
								 {"r3", BUTTON3_RELEASED},
								 {"c3", BUTTON3_CLICKED},
								 {"d3", BUTTON3_DOUBLE_CLICKED},
								 {"t3", BUTTON3_TRIPLE_CLICKED},
								 {"p4", BUTTON4_PRESSED},
								 {"r4", BUTTON4_RELEASED},
								 {"c4", BUTTON4_CLICKED},
								 {"d4", BUTTON4_DOUBLE_CLICKED},
								 {"t4", BUTTON4_TRIPLE_CLICKED},
								 {"C", BUTTON_CTRL},
								 {"S", BUTTON_SHIFT},
								 {"A", BUTTON_ALT},
								 {"m", REPORT_MOUSE_POSITION},
								 {NULL, 0}
	};
	int_table *iter;
	dstring *newd = ds_create(NULL);
	if(mask == 0)
		return newd;
	if(mask == ALL_MOUSE_EVENTS)
		return ds_appendstr(newd, "all");
	for(iter = events; iter->name; iter++)
	{
		if(mask & iter->val)
			ds_appendstr(newd, iter->name);
	}
	return newd;
}
#endif

//// curses.mousemask([mask])
//// if \em[mask] is present, sets the current terminal mouse event mask to \em[mask]
////
//// The mask has the following form:
//// \indent [+|-|=] (modifier | event-type button-no)*
//// \par \em[or] all \em[or] *
//// \noindent where \var[modifier] is one of
//// \item[p] -- a button is pressed
///  \item[r] -- a button is released
//// \item[c] -- a button is clicked
//// \item[d] -- a button is double clicked
//// \item[t] -- a button is triple clicked
//// \endlist
//// and the modifier is one of:
//// \item[C] -- a Control key is pressed 
//// \item[S] -- a Shift key is pressed
//// \item[A] -- an Alt key is pressed
//// \item[m] -- report mouse motions, not just button state change
//// \endlist
//// There are at most four buttons recognized.
//// 
//// \em[all] and \em[*] are equivalent to \em[p1p2p3p4r1r2r3r4c1c2c3c4d1d2d3d4t1t3t4CSA]
//// 
//// If a \var[mask] starts with \em[+], a new mask is a logical OR of the old mask and
/// \var[mask]; if \var[mask] starts with a \em[-], a new mask is logical AND of the old
/// mask and \var[mask] complement; otherwise, the mask is set to \var[mask].
//// \returns the old mask
//// \note this function is always present, even if the library doesn't support mice
static void bi_mousemask(chunk *current, interpreter *i, value *args)
{
#ifdef NCURSES_MOUSE_VERSION
	mmask_t saved;
	mousemask(0, &saved);
	if(!args)
		mousemask(saved, NULL);
	else
	{
		enum mask_mode { msk_set, msk_add, msk_sub };
		char *mstr = DS_BODY(args->val);
		int mode = msk_set;
		mmask_t newmask;
		switch(*mstr)
		{
		case '+':
			mode = msk_add;
			mstr++;
			break;
		case '-':
			mode = msk_sub;
			mstr++;
			break;
		case '=':
			mstr++;
		}
		newmask = convert_mouse_mask(mstr);
		if(mode == msk_add)
			newmask |= saved;
		else if(mode == msk_sub)
			newmask = (saved & ~newmask & ALL_MOUSE_EVENTS & REPORT_MOUSE_POSITION);
		mousemask(newmask, NULL);
	}
	ds_append(get_pstack()->state->buffer, stringize_mouse_mask(saved));
#endif
}

static master_instruction_src ncurses_src[] = {
								{"feature", 1, bi_feature},
								{"sysvar", 1, bi_sysvar},
								{NULL}
								};
										
static master_instruction_src ncurses_src_add[] = {
	{"curses.mousemask", 1, bi_mousemask, 0, 1, 1},
										{NULL}
										};

static int ncurses_inited;
void ncurses_LTX_finish(void)
{
	if(ncurses_inited)
		endwin();
}

//// \section[WINDOW STREAM]
static stream *win_open(interpreter *i, char *name, char *arg, int writable);

static void win_close(interpreter *i, stream *s)
{
	delwin(s->data);
}


//// Reading
//// ASCII characters read from the terminal are returned as is. Special keys are
//// read as \em[KEY] tag which has two attributes: \em[NAME] holding the name of the key
//// or \em[UNKNOWN] if the key is not recognized; and \em[CODE] holding the internal key
//// of the code.
/////
//// Key names are those defined by \ref[curses(3)] without the leading KEY_ (but
//// functional key names are just \em[Fnumber] without parentheses
/////
//// If the mouse is present, supported and enabled, mouse events are read as MOUSE tag
//// which has the following attributes:
//// \item[STATE] state of buttons in the same form as \em[mousemask]
//// \item[SCREENX] horizontal position of the mouse pointer relative to screen
//// \item[SCREENY] vertical position -"-
///  \item[X] horizontal position of the mouse pointer relative to the window;
//// not present if the pointer is outside the window
/// \item[Y] vertical position -"-
//// \item[Z] a Z coordinate of the pointer (currently always 0)
//// \endlist
static const char *key_to_name(int key);
static html_chunk *win_read(interpreter *i, stream *s)
{
	static html_chunk res;
	int ch = wgetch(s->data);
	if(ch == ERR) return NULL;
	release_html_chunk(&res);
	if(ch <= UCHAR_MAX)
		set_html_char(&res, ch);
#ifdef NCURSES_MOUSE_VERSION
	else if(ch == KEY_MOUSE)
	{
		int visible;
		MEVENT event;
		set_tag_name(&res, ds_fix(ds_create("MOUSE")));
		visible = getmouse(&event);
		set_html_param(&res, ds_create("STATE"), stringize_mouse_mask(event.bstate));
		set_html_param(&res, ds_create("SCREENX"), ds_fromint(event.x, 10, 0));
		set_html_param(&res, ds_create("SCREENY"), ds_fromint(event.y, 10, 0));
		set_html_param(&res, ds_create("Z"), ds_fromint(event.z, 10, 0));
		if(visible == OK && wmouse_trafo(s->data, &event.y, &event.x, FALSE))
		{
			set_html_param(&res, ds_create("X"), ds_fromint(event.x, 10, 0));
			set_html_param(&res, ds_create("Y"), ds_fromint(event.y, 10, 0));
		}
	}
#endif
	else
	{
		set_tag_name(&res, ds_fix(ds_create("KEY")));
		set_html_param(&res, ds_create("NAME"), ds_create(key_to_name(ch)));
		set_html_param(&res, ds_create("CODE"), ds_fromint(ch, 10, 0));
	}
	return &res;
}
	
static void win_writestr(interpreter *i, stream *s, char *str)
{
	waddstr(s->data, str);
}

static int lookup_int_table(const char *val, int_table *table)
{
	for(; table->name; table++)
	{
		if(val && !strcasecmp(val, table->name))
			return table->val;
	}
	return table->val;
}

static const char *rlookup_int_table(int val, int_table *table)
{
	for(; table->name; table++)
	{
		if(table->val == val)
			return table->name;
	}
	return NULL;
}

static int map_acs_symbol(const char *name)
{
	int_table acs_table[] = {
		{"ULCORNER", ACS_ULCORNER},
		{"LLCORNER", ACS_LLCORNER},
		{"URCORNER", ACS_URCORNER},
		{"LRCORNER", ACS_LRCORNER},
		{"LTEE", ACS_LTEE},
		{"RTEE", ACS_RTEE},
		{"BTEE", ACS_BTEE},
		{"TTEE", ACS_TTEE},
		{"HLINE", ACS_HLINE},
		{"VLINE", ACS_VLINE},
		{"PLUS", ACS_PLUS},
		{"S1", ACS_S1},
		{"S9", ACS_S9},
		{"DIAMOND", ACS_DIAMOND},
		{"CKBOARD", ACS_CKBOARD},
		{"DEGREE", ACS_DEGREE},
		{"PLMINUS", ACS_PLMINUS},
		{"BULLET", ACS_BULLET},
		{"LARROW", ACS_LARROW},
		{"RARROW", ACS_RARROW},
		{"DARROW", ACS_DARROW},
		{"UARROW", ACS_UARROW},
		{"BOARD", ACS_BOARD},
		{"LANTERN", ACS_LANTERN},
		{"BLOCK", ACS_BLOCK},
#ifdef ACS_S3
		{"S3", ACS_S3},
#else
		{"S3", '-'},
#endif
#ifdef ACS_S7
		{"S7", ACS_S7},
#else
		{"S7", '-'},
#endif
#ifdef ACS_LEQUAL
		{"LEQUAL", ACS_LEQUAL},
#else
		{"LEQUAL", '<'},
#endif
#ifdef ACS_GEQUAL
		{"GEQUAL", ACS_GEQUAL},
#else
		{"GEQUAL", '>'},
#endif
#ifdef ACS_PI
		{"PI", ACS_PI},
#else
		{"PI", '*'},
#endif
#ifdef ACS_NEQUAL
		{"NEQUAL", ACS_NEQUAL},
#else
		{"NEQUAL", '!'},
#endif
#ifdef ACS_STERLING
		{"STERLING", ACS_STERLING},
#else
		{"STERLING", 'f'},
#endif
		{NULL, 0}};
	return lookup_int_table(name, acs_table);
}

static int_table keys[] = {
	{"BREAK", KEY_BREAK},
	{"SRESET", KEY_SRESET},
	{"RESET", KEY_RESET},
	{"DOWN", KEY_DOWN},
	{"UP", KEY_UP},
	{"LEFT", KEY_LEFT},
	{"RIGHT", KEY_RIGHT},
	{"HOME", KEY_HOME},
	{"BACKSPACE", KEY_BACKSPACE},
	{"DL", KEY_DL},
	{"IL", KEY_IL},
	{"DC", KEY_DC},
	{"IC", KEY_IC},
	{"EIC", KEY_EIC},
	{"CLEAR", KEY_CLEAR},
	{"EOS", KEY_EOS},
	{"EOL", KEY_EOL},
	{"SF", KEY_SF},
	{"SR", KEY_SR},
	{"NPAGE", KEY_NPAGE},
	{"PPAGE", KEY_PPAGE},
	{"STAB", KEY_STAB},
	{"CTAB", KEY_CTAB},
	{"CATAB", KEY_CATAB},
	{"ENTER", KEY_ENTER},
	{"PRINT", KEY_PRINT},
	{"LL", KEY_LL},
	{"A1", KEY_A1},
	{"A3", KEY_A3},
	{"B2", KEY_B2},
	{"C1", KEY_C1},
	{"C3", KEY_C3},
	{"BTAB", KEY_BTAB},
	{"BEG", KEY_BEG},
	{"CANCEL", KEY_CANCEL},
	{"CLOSE", KEY_CLOSE},
	{"COMMAND", KEY_COMMAND},
	{"COPY", KEY_COPY},
	{"CREATE", KEY_CREATE},
	{"END", KEY_END},
	{"EXIT", KEY_EXIT},
	{"FIND", KEY_FIND},
	{"HELP", KEY_HELP},
	{"MARK", KEY_MARK},
	{"MESSAGE", KEY_MESSAGE},
	{"MOVE", KEY_MOVE},
	{"NEXT", KEY_NEXT},
	{"OPEN", KEY_OPEN},
	{"OPTIONS", KEY_OPTIONS},
	{"PREVIOUS", KEY_PREVIOUS},
	{"REDO", KEY_REDO},
	{"REFERENCE", KEY_REFERENCE},
	{"REFRESH", KEY_REFRESH},
	{"REPLACE", KEY_REPLACE},
	{"RESTART", KEY_RESTART},
	{"RESUME", KEY_RESUME},
	{"SAVE", KEY_SAVE},
	{"SBEG", KEY_SBEG},
	{"SCANCEL", KEY_SCANCEL},
	{"SCOMMAND", KEY_SCOMMAND},
	{"SCOPY", KEY_SCOPY},
	{"SCREATE", KEY_SCREATE},
	{"SDC", KEY_SDC},
	{"SDL", KEY_SDL},
	{"SELECT", KEY_SELECT},
	{"SEND", KEY_SEND},
	{"SEOL", KEY_SEOL},
	{"SEXIT", KEY_SEXIT},
	{"SFIND", KEY_SFIND},
	{"SHELP", KEY_SHELP},
	{"SHOME", KEY_SHOME},
	{"SIC", KEY_SIC},
	{"SLEFT", KEY_SLEFT},
	{"SMESSAGE", KEY_SMESSAGE},
	{"SMOVE", KEY_SMOVE},
	{"SNEXT", KEY_SNEXT},
	{"SOPTIONS", KEY_SOPTIONS},
	{"SPREVIOUS", KEY_SPREVIOUS},
	{"SPRINT", KEY_SPRINT},
	{"SREDO", KEY_SREDO},
	{"SREPLACE", KEY_SREPLACE},
	{"SRIGHT", KEY_SRIGHT},
	{"SRSUME", KEY_SRSUME},
	{"SSAVE", KEY_SSAVE},
	{"SSUSPEND", KEY_SSUSPEND},
	{"SUNDO", KEY_SUNDO},
	{"SUSPEND", KEY_SUSPEND},
	{"UNDO", KEY_UNDO},
	{NULL, 0}
};

static const char *key_to_name(int key)
{
	static char buf[8];
	const char *name;
	if(key >= KEY_F(0) && key <= KEY_F(63))
	{
		sprintf(buf, "F%d", key - KEY_F0);
		return buf;
	}
	name = rlookup_int_table(key, keys);
	return name ? name : "UNKNOWN";
}

static int name_to_key(const char *name)
{
	if((*name == 'F' || *name == 'f') && isdigit(name[1]))
		return KEY_F0 + strtoul(name + 1, NULL, 10);
	return lookup_int_table(name, keys);
}

//// Writing
//// ASCII characters are written as is.
//// The following tags are recognized:
//// \item[BORDER] draws a border around the window
//// \item[CLEAR] clear the window
//// \item[MOVE] moves the cursor to \em[X], \em[Y]; or offset to \em[DX], \em[DY]
/// \item[HLINE] draws a horizontal line from the cursor of the length \em[LEN]
//// \item[VLINE] draws a vertical line -"-
//// \item[STANDOUT] turns on standout hiliting
//// \item[/STANDOUT] turns it off
//// \item[BOLD] turns on bold hiliting
//// \item[/BOLD] turns it off
//// \item[&] writes a special character whose name is \em[VALUE]
//// Character names are the same as defined by \ref[curses(3)] without the leading ACS_
//// In addition, \em[LEQUAL], \em[GEQUAL], \em[NEQUAL], \em[S3], \em[S7], \em[PI] and
//// \em[STERLING] are ASCII approximated if the library lacks them
static void win_writetag(interpreter *i, stream *s, html_chunk *ch)
{
	if(ch->type == hct_char)
		waddch(s->data, get_html_char(ch));
	else if(ch->type == hct_tag)
	{
		char *name = get_tag_name(ch);
		if(!strcasecmp(name, "BORDER"))
			box((WINDOW *)s->data, 0, 0);
		else if(!strcasecmp(name, "CLEAR"))
			wclear(s->data);
		else if(!strcasecmp(name, "MOVE"))
		{
			int x, y;
			char *tmp;
			getyx((WINDOW *)s->data, y, x);
			tmp = get_html_param(ch, "X");
			if(tmp)
				x = strtoul(tmp, NULL, 10);
			else if((tmp = get_html_param(ch, "DX")))
				x += strtol(tmp, NULL, 10);
			tmp = get_html_param(ch, "Y");
			if(tmp)
				y = strtoul(tmp, NULL, 10);
			else if((tmp = get_html_param(ch, "DY")))
				y += strtol(tmp, NULL, 10);	
			wmove((WINDOW *)s->data, y, x);
		}
		else if(!strcasecmp(name, "HLINE"))
		{
			char *tmp = get_html_param(ch, "LEN");
			if(!tmp)
				recover_error("ncurses: HLINE must have LEN");
			whline((WINDOW *)s->data, ACS_HLINE, strtoul(tmp, NULL, 10));
		}
		else if(!strcasecmp(name, "VLINE"))
		{
			char *tmp = get_html_param(ch, "LEN");
			if(!tmp)
				recover_error("ncurses: VLINE must have LEN");
			wvline((WINDOW *)s->data, ACS_VLINE, strtoul(tmp, NULL, 10));
		}
		else if(!strcasecmp(name, "STANDOUT"))
			wattron((WINDOW *)s->data, A_STANDOUT);
		else if(!strcasecmp(name, "/STANDOUT"))
			wattroff((WINDOW *)s->data, A_STANDOUT);
		else if(!strcasecmp(name, "BOLD"))
			wattron((WINDOW *)s->data, A_BOLD);
		else if(!strcasecmp(name, "/BOLD"))
			wattroff((WINDOW *)s->data, A_BOLD);
		else if(!strcmp(name, "&"))
		{
			int acs = map_acs_symbol(get_html_param(ch, "VALUE"));
			if(acs)
				waddch((WINDOW *)s->data, acs);
		}
	}	
}

//// Ioctls
/// The following ioctls are recognized:
//// \item[flush] actually draws the screen
//// \item[scrollok] enables scrolling of the window
//// \item[noscrollok] disables -"-
static char *win_ioctl(interpreter *i, stream *s, char *ctl)
{
	if(!strcasecmp(ctl, "flush"))
	{
		wrefresh(s->data);
		return "";
	}
	else if(!strcasecmp(ctl, "scrollok"))
	{
		scrollok((WINDOW *)s->data, TRUE);
		return "";
	}
	else if(!strcasecmp(ctl, "noscrollok"))
	{
		scrollok((WINDOW *)s->data, FALSE);
		return "";
	}
	return NULL;
}

//// Positioning
//// \em[getpos] on a window stream will the following value:
//// win_width * ypos + xpos
////
//// Resp. \em[setpos] will convert such number to a coordinate pair
////
//// \em[rewind] will move the cursor to the top left corner
static int win_getpos(interpreter *i, stream *s)
{
	int wdt, hgt, y, x;
	getmaxyx((WINDOW *)s->data, wdt, hgt);
	wdt++;
	getyx((WINDOW *)s->data, y, x);
	return y * wdt + x;
}

static void win_setpos(interpreter *i, stream *s, int pos)
{
	int wdt, hgt;
	getmaxyx((WINDOW *)s->data, wdt, hgt);
	wdt++;
	wmove((WINDOW *)s->data, pos / wdt, pos % wdt);
}

static void win_rewind(interpreter *i, stream *s)
{
	wmove(s->data, 0, 0);
}

static stream_class ncurses_stream_class = {"window", win_open, win_close,
											win_read, win_writestr, win_writetag,
											win_ioctl, win_getpos, win_setpos, win_rewind
										};

///// Crearing window streams
//// The type of window streams is \em[window]
//// The \em[arg] is X-like geometry specification:
//// \indent width x height [[+|-] xpos] [[+|-] ypos] \noindent
static stream *win_open(interpreter *i, char *name, char *arg, int writable)
{
	stream *ns = ALLOC(stream);
	WINDOW *nw;
	int wdt, hgt, x, y;
	ns->class = &ncurses_stream_class;
	ns->direction = 1;
	wdt = strtol(arg, &arg, 10);
	if(*arg == 'x')
		arg++;
	hgt = strtol(arg, &arg, 10);
	x = strtol(arg, &arg, 10);
	y = strtol(arg, &arg, 10);
	if(x < 0)
		x = COLS  - wdt + x + 1;
	if(y < 0)
		y = LINES - hgt + y + 1; 
	nw = newwin(hgt, wdt, y, x);
	if(!nw)
		recover_error("ncurses error: can't create a window");
	keypad(nw, TRUE);
	idlok(nw, TRUE);
	ns->data = nw;
	return ns;
}

void ncurses_LTX_attach(mod_attach *ma)
{
	if(!ncurses_inited)
	{
		initscr();
		if(has_colors())
			start_color();
		cbreak();
		noecho();
		nonl();
		intrflush(stdscr, FALSE);
		ncurses_inited = 1;
	}	
	register_mod_stream(ma->i, &ncurses_stream_class);
	ma->data = xmalloc(sizeof(ncurses_src) * sizeof(instruction_t) / sizeof(*ncurses_src));
	replace_builtins(ma->i, ncurses_src, ma->data);
	append_builtins(ma->i, ncurses_src_add);
}

const char ncurses_LTX_version[] = VERSION;
const char ncurses_LTX_date[] = __DATE__;
const char ncurses_LTX_srcdate[] = "$Id: ncurses.c,v 1.7 2002/12/09 10:41:40 aa5779 Exp $";


/// \trailer
